use strict;
use Module::Build;
my ($build, $params);


$params = {
            'requires' => {
                            'String::Util' => '1.24',
                            'File::Copy' => 0,
                            'Test::Toolbox' => '0.2',
                            'File::Copy::Recursive' => 0,
                            'File::Path' => 0,
                            'File::Find' => 0,
                            'File::chdir' => 0,
                            'File::Compare' => 0,
                            'Class::ISA' => 0
                          },
            'dist_version_from' => 'lib/File/Util.pm',
            'dist_author' => 'Miko O\'Sullivan <miko@idocs.com>',
            'dist_abstract' => 'File::Util -- handy file tools',
            'license' => 'perl',
            'module_name' => 'File::Util'
          };


$build = Module::Build->new(%$params);
$build->create_build_script();
